from .request import Request
from .response import Response
from .primitives import Int, Buffer, UString


AUTH_XID = -4


class AuthRequest(Request):
    """
    """
    opcode = 100
    special_xid = AUTH_XID

    parts = (
        ("type", Int),
        ("scheme", UString),
        ("auth", Buffer),
    )


class AuthResponse(Response):
    """
    """
    opcode = 100

    parts = ()



from aiozk import exc

from .base_lock import BaseLock


class Lock(BaseLock):

    async def acquire(self, timeout=None):
        result = None
        while not result:
            try:
                result = await self.wait_in_line("lock", timeout)
            except exc.SessionLost:
                continue
        return result



from .acl import (  # noqa
    GetACLRequest,
    GetACLResponse,
    SetACLRequest,
    SetACLResponse,
    WORLD_READABLE,
    AUTHED_UNRESTRICTED,
    UNRESTRICTED_ACCESS,
)
from .auth import (  # noqa
    AuthRequest,
    AuthResponse,
    AUTH_XID,
)
from .check import (  # noqa
    CheckVersionRequest,
    CheckVersionResponse,
)
from .children import (  # noqa
    GetChildrenRequest,
    GetChildrenResponse,
    GetChildren2Request,
    GetChildren2Response,
)
from .close import (  # noqa
    CloseRequest,
    CloseResponse,
    CLOSE_XID,
)
from .connect import (  # noqa
    ConnectRequest,
    ConnectResponse,
)
from .create import (  # noqa
    CreateRequest,
    CreateResponse,
    Create2Request,
    Create2Response,
)
from .data import (  # noqa
    GetDataRequest,
    GetDataResponse,
    SetDataRequest,
    SetDataResponse,
)
from .delete import (  # noqa
    DeleteRequest,
    DeleteResponse,
)
from .exists import (  # noqa
    ExistsRequest,
    ExistsResponse,
)
from .ping import (  # noqa
    PingRequest,
    PingResponse,
    PING_XID,
)
from .reconfig import (  # noqa
    ReconfigRequest,
    ReconfigResponse,
)
from .sasl import (  # noqa
    SASLRequest,
    SASLResponse,
)
from .sync import (  # noqa
    SyncRequest,
    SyncResponse,
)
from .watches import (  # noqa
    WatchEvent,
    SetWatchesRequest,
    SetWatchesResponse,
    CheckWatchesRequest,
    CheckWatchesResponse,
    RemoveWatchesRequest,
    RemoveWatchesResponse,
    WATCH_XID,
)
from .response import (  # noqa
    response_xref
)
from .transaction import (  # noqa
    TransactionRequest,
    TransactionResponse,
)

SPECIAL_XIDS = (AUTH_XID, PING_XID, CLOSE_XID)



version_info = (0, 5, 3)

__version__ = ".".join((str(point) for point in version_info))

from .client import ZKClient  # noqa
from .protocol import WatchEvent  # noqa
from .protocol.acl import ACL  # noqa
from .retry import RetryPolicy  # noqa



import logging
import struct
from io import BytesIO

from .part import Part
from .primitives import Int


log = logging.getLogger(__name__)


class Request(Part):
    """
    Returns a bytesring representation of the request instance.

    # TODO(wglass): specify how xid and type preamble goes in

    Since this is a ``Part`` subclass the rest is a matter of
    appending the result of a ``render()`` call.
    """
    opcode = None
    special_xid = None
    writes_data = False

    def serialize(self, xid=None):
        buff = BytesIO()

        formats = []
        data = []
        if xid is not None:
            formats.append(Int.fmt)
            data.append(xid)
        if self.opcode:
            formats.append(Int.fmt)
            data.append(self.opcode)

        payload_format, payload_data = self.render()
        formats.append(payload_format)
        data.extend(payload_data)

        buff.write(struct.pack("!" + "".join(formats), *data))

        return buff.getvalue()



import asyncio
import logging

from .children_watcher import ChildrenWatcher
from .data_watcher import DataWatcher
from .recipe import Recipe
from ..exc import NoNode



log = logging.getLogger(__name__)


class TreeCache(Recipe):

    sub_recipes = {
        "data_watcher": DataWatcher,
        "child_watcher": ChildrenWatcher,
    }

    def __init__(self, base_path, defaults=None):
        super(TreeCache, self).__init__(base_path)
        self.defaults = defaults or {}

        self.root = None

    async def start(self):
        log.debug("Starting znode tree cache at %s", self.base_path)

        self.root = ZNodeCache(
            self.base_path, self.defaults,
            self.client, self.data_watcher, self.child_watcher,
        )

        await self.ensure_path()

        await self.root.start()

    async def stop(self):
        await self.root.stop()

    def __getattr__(self, attribute):
        return getattr(self.root, attribute)

    def as_dict(self):
        return self.root.as_dict()


class ZNodeCache(object):

    def __init__(self, path, defaults, client, data_watcher, child_watcher):
        self.path = path

        self.client = client
        self.defaults = defaults

        self.data_watcher = data_watcher
        self.child_watcher = child_watcher

        self.children = {}
        self.data = None

    @property
    def dot_path(self):
        return self.path[1:].replace("/", ".")

    @property
    def value(self):
        return self.data

    def __getattr__(self, name):
        if name not in self.children:
            raise AttributeError

        return self.children[name]

    async def start(self):
        self.data = await self.client.get_data(self.path)
        for child in await self.client.get_children(self.path):
            self.children[child] = ZNodeCache(
                self.path + "/" + child, self.defaults.get(child, {}),
                self.client, self.data_watcher, self.child_watcher
            )

        await asyncio.gather(*(child.start() for child in self.children.values()), loop=self.client.loop)

        self.data_watcher.add_callback(self.path, self.data_callback)
        self.child_watcher.add_callback(self.path, self.child_callback)

    async def stop(self):
        await asyncio.sleep(0.02)
        await asyncio.gather(*(child.stop() for child in self.children.values()), loop=self.client.loop)
        self.data_watcher.remove_callback(self.path, self.data_callback)
        self.child_watcher.remove_callback(self.path, self.child_callback)

    async def child_callback(self, new_children):
        if new_children == NoNode:
            return
        removed_children = set(self.children.keys()) - set(new_children)
        added_children = set(new_children) - set(self.children.keys())

        for removed in removed_children:
            log.debug("Removed child %s", self.dot_path + "." + removed)
            child = self.children.pop(removed)
            await child.stop()

        for added in added_children:
            log.debug("Added child %s", self.dot_path + "." + added)
            self.children[added] = ZNodeCache(
                self.path + "/" + added, self.defaults.get(added, {}),
                self.client, self.data_watcher, self.child_watcher
            )
        await asyncio.gather(*(self.children[added].start() for added in added_children), loop=self.client.loop)

    async def data_callback(self, data):
        log.debug("New value for %s: %r", self.dot_path, data)
        if data == NoNode:
            return
        self.data = data

    def as_dict(self):
        if self.children:
            return {
                child_path: child_znode.as_dict()
                for child_path, child_znode in self.children.items()
            }
        return self.data



from .request import Request
from .response import Response
from .primitives import Buffer


class SASLRequest(Request):
    """
    """
    opcode = 102

    parts = (
        ("token", Buffer)
    )


class SASLResponse(Response):
    """
    """
    opcode = 102

    parts = (
        ("token", Buffer)
    )



class ZKError(Exception):
    pass


class ConnectError(ZKError):
    def __init__(self, host, port, server_id=None):
        self.host = host
        self.port = port
        self.server_id = server_id

    def __str__(self):
        return "Error connecting to %s:%s" % (self.host, self.port)


class NoServersError(ZKError):
    pass


class SessionLost(ZKError):
    pass


class InvalidClientState(ZKError):
    pass


class TimeoutError(ZKError):
    pass


class UnfinishedRead(ZKError):
    pass


class FailedRetry(ZKError):
    pass


response_error_xref = {}


class ResponseErrorMeta(type):

    def __new__(cls, name, bases, attrs):
        new_class = super(ResponseErrorMeta, cls).__new__(
            cls, name, bases, attrs
        )

        response_error_xref[new_class.error_code] = new_class

        return new_class


class ResponseError(ZKError, metaclass=ResponseErrorMeta):
    error_code = None

    def __str__(self):
        return self.__class__.__name__


class UnknownError(ResponseError):

    def __init__(self, error_code):
        print(repr(error_code))
        self.error_code = error_code

    def __str__(self):
        return "Unknown error code: %s" % self.error_code


def get_response_error(error_code):
    if error_code not in response_error_xref:
        return UnknownError(error_code)

    return response_error_xref[error_code]()


class RolledBack(ResponseError):
    error_code = 0


class ZKSystemError(ResponseError):
    error_code = -1


class RuntimeInconsistency(ResponseError):
    error_code = -2


class DataInconsistency(ResponseError):
    error_code = -3


class ConnectionLoss(ResponseError):
    error_code = -4


class MarshallingError(ResponseError):
    error_code = -5


class Unimplemented(ResponseError):
    error_code = -6


class OperationTimeout(ResponseError):
    error_code = -7


class BadArguments(ResponseError):
    error_code = -8


class UnknownSession(ResponseError):
    error_code = -12


class NewConfigNoQuorum(ResponseError):
    error_code = -13


class ReconfigInProcess(ResponseError):
    error_code = -14


class APIError(ResponseError):
    error_code = -100


class NoNode(ResponseError):
    error_code = -101


class NoAuth(ResponseError):
    error_code = -102


class BadVersion(ResponseError):
    error_code = -103


class NoChildrenForEphemerals(ResponseError):
    error_code = -108


class NodeExists(ResponseError):
    error_code = -110


class NotEmpty(ResponseError):
    error_code = -111


class SessionExpired(ResponseError):
    error_code = -112


class InvalidCallback(ResponseError):
    error_code = -113


class InvalidACL(ResponseError):
    error_code = -114


class AuthFailed(ResponseError):
    error_code = -115


class SessionMoved(ResponseError):
    error_code = -118


class NotReadOnly(ResponseError):
    error_code = -119


class EphemeralOnLocalSession(ResponseError):
    error_code = -120


class NoWatcher(ResponseError):
    error_code = -121



from .part import Part


response_xref = {}


class ResponseMeta(type):

    def __new__(cls, name, bases, attrs):
        new_class = super(ResponseMeta, cls).__new__(cls, name, bases, attrs)

        response_xref[new_class.opcode] = new_class

        return new_class


class Response(Part, metaclass=ResponseMeta):
    """
    Base class for all operation response classes.

    A simple class, has only an ``opcode`` attribute expected to be defined by
    subclasses, and a `deserialize()` classmethod.
    """
    opcode = None

    @classmethod
    def deserialize(cls, raw_bytes):
        """
        Deserializes the given raw bytes into an instance.

        Since this is a subclass of ``Part`` but a top-level one (i.e. no other
        subclass of ``Part`` would have a ``Response`` as a part) this merely
        has to parse the raw bytes and discard the resulting offset.
        """
        instance, _ = cls.parse(raw_bytes, offset=0)

        return instance



from .request import Request
from .response import Response
from .primitives import Int, UString


class CheckVersionRequest(Request):
    """
    """
    opcode = 13

    parts = (
        ("path", UString),
        ("version", Int),
    )


class CheckVersionResponse(Response):
    """
    """
    opcode = 13

    parts = ()



from aiozk import WatchEvent

from .base_watcher import BaseWatcher


class ChildrenWatcher(BaseWatcher):

    watched_events = [WatchEvent.CHILDREN_CHANGED, WatchEvent.DELETED]

    async def fetch(self, path):
        children = await self.client.get_children(path=path, watch=True)
        return children
